@@ -277,7 +277,6 @@ fn windows_python_info(executable: &Path) -> Result<Option<InterpreterConfig>> {
             interpreter_kind: InterpreterKind::CPython,
             abiflags: String::new(),
             ext_suffix: String::new(),
-            abi_tag: None,
             pointer_width: Some(pointer_width),
         }))
     } else {
@@ -291,6 +290,7 @@ fn windows_python_info(executable: &Path) -> Result<Option<InterpreterConfig>> {
 pub enum InterpreterKind {
     CPython,
     PyPy,
+    GraalPy,
 }
 
 impl InterpreterKind {
@@ -303,6 +303,11 @@ impl InterpreterKind {
     pub fn is_pypy(&self) -> bool {
         matches!(self, InterpreterKind::PyPy)
     }
+
+    /// Is this a GraalPy interpreter?
+    pub fn is_graalpy(&self) -> bool {
+        matches!(self, InterpreterKind::GraalPy)
+    }
 }
 
 impl fmt::Display for InterpreterKind {
@@ -310,6 +315,7 @@ impl fmt::Display for InterpreterKind {
         match *self {
             InterpreterKind::CPython => write!(f, "CPython"),
             InterpreterKind::PyPy => write!(f, "PyPy"),
+            InterpreterKind::GraalPy => write!(f, "GraalPy"),
         }
     }
 }
@@ -321,6 +327,7 @@ impl FromStr for InterpreterKind {
         match s.to_ascii_lowercase().as_str() {
             "cpython" => Ok(InterpreterKind::CPython),
             "pypy" => Ok(InterpreterKind::PyPy),
+            "graalvm" | "graalpy" => Ok(InterpreterKind::GraalPy),
             unknown => Err(format!("Unknown interpreter kind '{unknown}'")),
         }
     }
@@ -341,7 +348,6 @@ struct InterpreterMetadataMessage {
     // comes from `platform.system()`
     system: String,
     soabi: Option<String>,
-    abi_tag: Option<String>,
 }
 
 /// The location and version of an interpreter
@@ -399,16 +405,16 @@ fn fun_with_abiflags(
         )
     }
 
-    if message.major != 3 || message.minor < 5 {
+    if message.major != 3 || message.minor < 7 {
         bail!(
-            "Only python >= 3.5 is supported, while you're using python {}.{}",
+            "Only python >= 3.7 is supported, while you're using python {}.{}",
             message.major,
             message.minor
         );
     }
 
-    if message.interpreter == "pypy" {
-        // pypy does not specify abi flags
+    if message.interpreter == "pypy" || message.interpreter == "graalvm" {
+        // pypy and graalpy do not specify abi flags
         Ok("".to_string())
     } else if message.system == "windows" {
         if matches!(message.abiflags.as_deref(), Some("") | None) {
@@ -438,7 +444,7 @@ impl PythonInterpreter {
         } else {
             match self.interpreter_kind {
                 InterpreterKind::CPython => true,
-                InterpreterKind::PyPy => false,
+                InterpreterKind::PyPy | InterpreterKind::GraalPy => false,
             }
         }
     }
@@ -505,17 +511,29 @@ impl PythonInterpreter {
                     // pypy uses its version as part of the ABI, e.g.
                     // pypy 3.7 7.3 => numpy-1.20.1-pp37-pypy37_pp73-manylinux2014_x86_64.whl
                     format!(
-                        "pp{major}{minor}-pypy{major}{minor}_{abi_tag}-{platform}",
+                        "pp{major}{minor}-{abi_tag}-{platform}",
                         major = self.major,
                         minor = self.minor,
-                        // TODO: Proper tag handling for pypy
-                        abi_tag = self
-                            .abi_tag
-                            .clone()
-                            .expect("PyPy's syconfig didn't define an `SOABI` ಠ_ಠ"),
+                        abi_tag = calculate_abi_tag(&self.ext_suffix)
+                            .expect("PyPy's syconfig didn't define a valid `EXT_SUFFIX` ಠ_ಠ"),
                         platform = platform,
                     )
                 }
+                InterpreterKind::GraalPy => {
+                    // GraalPy suffers from pypa/packaging#614, where
+                    // packaging misdetects it as a 32-bit implementation,
+                    // so GraalPy adds the correct platform itself, e.g.
+                    // graalpy 3.10 23.1 => numpy-1.23.5-graalpy310-graalpy231_310_native_x86_64_linux-linux_i686.whl
+                    format!(
+                        "graalpy{major}{minor}-{abi_tag}_{arch}_{os}-{os}_i686",
+                        major = self.major,
+                        minor = self.minor,
+                        abi_tag = calculate_abi_tag(&self.ext_suffix)
+                            .expect("GraalPy's syconfig didn't define a valid `EXT_SUFFIX` ಠ_ಠ"),
+                        os = target.get_python_os(),
+                        arch = target.get_python_arch(),
+                    )
+                }
             }
         };
         Ok(tag)
@@ -635,6 +653,7 @@ impl PythonInterpreter {
         let interpreter = match message.interpreter.as_str() {
             "cpython" => InterpreterKind::CPython,
             "pypy" => InterpreterKind::PyPy,
+            "graalvm" | "graalpy" => InterpreterKind::GraalPy,
             other => {
                 bail!("Unsupported interpreter {}", other);
             }
@@ -670,7 +689,6 @@ impl PythonInterpreter {
                 ext_suffix: message
                     .ext_suffix
                     .context("syconfig didn't define an `EXT_SUFFIX` ಠ_ಠ")?,
-                abi_tag: message.abi_tag,
                 pointer_width: None,
             },
             executable,
@@ -888,7 +906,7 @@ impl PythonInterpreter {
     pub fn get_venv_site_package(&self, venv_base: impl AsRef<Path>, target: &Target) -> PathBuf {
         if target.is_unix() {
             match self.interpreter_kind {
-                InterpreterKind::CPython => {
+                InterpreterKind::CPython | InterpreterKind::GraalPy => {
                     let python_dir = format!("python{}.{}", self.major, self.minor);
                     venv_base
                         .as_ref()
@@ -929,6 +947,35 @@ impl fmt::Display for PythonInterpreter {
     }
 }
 
+/// Calculate the ABI tag from EXT_SUFFIX
+fn calculate_abi_tag(ext_suffix: &str) -> Option<String> {
+    let parts = ext_suffix.split('.').collect::<Vec<_>>();
+    if parts.len() < 3 {
+        // CPython3.7 and earlier uses ".pyd" on Windows.
+        return None;
+    }
+    let soabi = parts[1];
+    let mut soabi_split = soabi.split('-');
+    let abi = if soabi.starts_with("cpython") {
+        // non-windows
+        format!("cp{}", soabi_split.nth(1).unwrap())
+    } else if soabi.starts_with("cp") {
+        // windows
+        soabi_split.next().unwrap().to_string()
+    } else if soabi.starts_with("pypy") {
+        soabi_split.take(2).collect::<Vec<_>>().join("-")
+    } else if soabi.starts_with("graalpy") {
+        soabi_split.take(3).collect::<Vec<_>>().join("-")
+    } else if !soabi.is_empty() {
+        // pyston, ironpython, others?
+        soabi_split.nth(1).unwrap().to_string()
+    } else {
+        return None;
+    };
+    let abi_tag = abi.replace(['.', '-', ' '], "_");
+    Some(abi_tag)
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -952,4 +999,28 @@ mod tests {
         );
         assert_eq!(pythons.len(), 4);
     }
+
+    #[test]
+    fn test_calculate_abi_tag() {
+        let cases = vec![
+            (".cpython-37m-x86_64-linux-gnu.so", Some("cp37m")),
+            (".cpython-310-x86_64-linux-gnu.so", Some("cp310")),
+            (".cpython-310-darwin.so", Some("cp310")),
+            (".cp310-win_amd64.pyd", Some("cp310")),
+            (".cp39-mingw_x86_64.pyd", Some("cp39")),
+            (".cpython-312-wasm32-wasi.so", Some("cp312")),
+            (".cpython-38.so", Some("cp38")),
+            (".pyd", None),
+            (".so", None),
+            (".pypy38-pp73-x86_64-linux-gnu.so", Some("pypy38_pp73")),
+            (
+                ".graalpy-38-native-x86_64-darwin.dylib",
+                Some("graalpy_38_native"),
+            ),
+            (".pyston-23-x86_64-linux-gnu.so", Some("23")),
+        ];
+        for (ext_suffix, expected) in cases {
+            assert_eq!(calculate_abi_tag(ext_suffix).as_deref(), expected);
+        }
+    }
 }
